from ngsolve import *
from netgen.geom2d import unit_square
from math import pi

G = sin(pi * x)
H = -G
uex = (1 - y) * G 
gradu = (pi * (1 - y) * cos(pi * x), H)
F = pi * pi * uex

order_flux = 2
mesh = Mesh(unit_square.GenerateMesh(maxh = 0.1))
V = HDiv(mesh, order = order_flux, dirichlet = 'top')
Q = L2(mesh, order = order_flux - 1, dirichlet = 'left|bottom|right')

sigma, tau = V.TnT()
u, v = Q.TnT()

normal = specialcf.normal(mesh.dim)

a = BilinearForm(V)
a += sigma * tau * dx
a.Assemble()

b = BilinearForm(trialspace = V, testspace = Q)
b += div(sigma) * v * dx
b.Assemble()

f = LinearForm(V)
f += G * (tau.Trace() * normal) * ds \
(definedon = mesh.Boundaries('left|bottom|right'))
f.Assemble()

g = LinearForm(Q)
g += -F * v  * dx
g.Assemble()

gfsigma = GridFunction(V, 'sigma')
gfsigma.Set(H * normal, BND)
gfu = GridFunction(Q, 'u')

K = BlockMatrix( [ [a.mat, b.mat.T], [b.mat, None] ] )

bT = b.mat.CreateTranspose()
D = la.SparseMatrixd.CreateFromCOO(indi = [k for k in range(a.mat.height)], indj = [k for k in range(a.mat.height)],
                                   values = [1.0/a.mat[k,k] for k in range(a.mat.height)], h = a.mat.height, w = a.mat.height)
M = b.mat @ D @ bT
M_jac = M.CreateSmoother()

rhs = BlockVector( [f.vec, g.vec] )
sol = BlockVector( [gfsigma.vec, gfu.vec] )
pre = BlockMatrix( [ [D, None], [None, M_jac] ] )
# pre = None

solvers.MinRes(mat = K, rhs = rhs, sol = sol, pre=pre, \
               maxsteps = 5000, initialize = False) 

Draw(gfu)

gfuex = GridFunction(Q)
gfuex.Set(uex)

gfsigmaex = GridFunction(V)
gfsigmaex.Set(gradu)

print('u error: ', sqrt(Integrate((gfuex - gfu) ** 2, mesh)))
print('sigma error: ', sqrt(Integrate((gfsigmaex - gfsigma) ** 2, mesh)))


